@@ -13,25 +13,26 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import base64
 import gzip
 import io
 import logging
 import math
 import os
 import re
-import shlex
 import shutil
 import stat
 import tempfile
 import time
+import warnings
 
 from oslo_concurrency import processutils
 from oslo_config import cfg
-from oslo_serialization import base64
 from oslo_utils import excutils
 from oslo_utils import imageutils
 from oslo_utils import units
 import requests
+import tenacity
 
 from ironic_lib.common.i18n import _
 from ironic_lib import disk_partitioner
@@ -59,6 +60,13 @@ opts = [
                default=10,
                help='Maximum number of attempts to try to read the '
                     'partition.'),
+    cfg.IntOpt('image_convert_memory_limit',
+               default=1024,
+               help='Memory limit for "qemu-img convert" in MiB. Implemented '
+                    'via the address space resource limit.'),
+    cfg.IntOpt('image_convert_attempts',
+               default=3,
+               help='Number of attempts to convert an image.'),
 ]
 
 CONF = cfg.CONF
@@ -72,11 +80,22 @@ _PARTED_PRINT_RE = re.compile(r"^(\d+):([\d\.]+)MiB:"
 CONFIGDRIVE_LABEL = "config-2"
 MAX_CONFIG_DRIVE_SIZE_MB = 64
 
+GPT_SIZE_SECTORS = 33
+
 # Maximum disk size supported by MBR is 2TB (2 * 1024 * 1024 MB)
 MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR = 2097152
 
 # Limit the memory address space to 1 GiB when running qemu-img
-QEMU_IMG_LIMITS = processutils.ProcessLimits(address_space=1 * units.Gi)
+QEMU_IMG_LIMITS = None
+
+
+def _qemu_img_limits():
+    global QEMU_IMG_LIMITS
+    if QEMU_IMG_LIMITS is None:
+        QEMU_IMG_LIMITS = processutils.ProcessLimits(
+            address_space=CONF.disk_utils.image_convert_memory_limit
+            * units.Mi)
+    return QEMU_IMG_LIMITS
 
 
 def list_partitions(device):
@@ -85,7 +104,7 @@ def list_partitions(device):
     :param device: The device path.
     :returns: list of dictionaries (one per partition) with keys:
               number, start, end, size (in MiB), filesystem, partition_name,
-              flags
+              flags, path.
     """
     output = utils.execute(
         'parted', '-s', '-m', device, 'unit', 'MiB', 'print',
@@ -108,7 +127,9 @@ def list_partitions(device):
         # Cast int fields to ints (some are floats and we round them down)
         groups = [int(float(x)) if i < 4 else x
                   for i, x in enumerate(match.groups())]
-        result.append(dict(zip(fields, groups)))
+        item = dict(zip(fields, groups))
+        item['path'] = partition_index_to_path(device, item['number'])
+        result.append(item)
     return result
 
 
@@ -151,27 +172,77 @@ def get_disk_identifier(dev):
     """
     disk_identifier = utils.execute('hexdump', '-s', '440', '-n', '4',
                                     '-e', '''\"0x%08x\"''',
-                                    dev,
-                                    run_as_root=True,
-                                    check_exit_code=[0],
-                                    attempts=5,
-                                    delay_on_retry=True)
+                                    dev, run_as_root=True,
+                                    attempts=5, delay_on_retry=True)
     return disk_identifier[0]
 
 
+def get_partition_table_type(device):
+    """Get partition table type, msdos or gpt.
+
+    :param device: the name of the device
+    :return: dos, gpt or None
+    """
+    return get_device_information(device, probe=True).get('PTTYPE')
+
+
+def get_device_information(device, probe=False, fields=None):
+    """Get information about a device using blkid.
+
+    Can be applied to all block devices: disks, RAID, partitions.
+
+    :param device: Device name.
+    :param probe: Switch to low-level probing mode.
+    :param fields: A list of fields to request (all by default).
+    :return: A dictionary with requested fields as keys.
+    :raises: ProcessExecutionError
+    """
+    args = []
+    if probe:
+        args.append('--probe')
+    if fields:
+        args += sum((['--match-tag', field] for field in fields), [])
+
+    output, err = utils.execute('blkid', device, *args,
+                                use_standard_locale=True, run_as_root=True)
+    if output.strip():
+        output = output.split(': ', 1)[1]
+        return next(utils.parse_device_tags(output))
+    else:
+        return {}
+
+
+def find_efi_partition(device):
+    """Looks for the EFI partition on a given device.
+
+    A boot partition on a GPT disk is assumed to be an EFI partition as well.
+
+    :param device: the name of the device
+    :return: the EFI partition record from `list_partitions` or None
+    """
+    is_gpt = get_partition_table_type(device) == 'gpt'
+    for part in list_partitions(device):
+        flags = {x.strip() for x in part['flags'].split(',')}
+        if 'esp' in flags or ('boot' in flags and is_gpt):
+            LOG.debug("Found EFI partition %s on device %s", part, device)
+            return part
+    else:
+        LOG.debug("No efi partition found on device %s", device)
+
+
 def get_uefi_disk_identifier(dev):
     """Get the uuid from the disk being exposed by the ramdisk.
 
-    This uuid is appended to the pxe config which will then be set as the root
-    and load the bootx64.efi file using chainloader and boot the machine.
-    This is helpful in deployments to nodes with multiple disks.
-
-    https://wiki.gentoo.org/wiki/GRUB2/Chainloading
+    DEPRECATED: use find_efi_partition with get_device_information instead.
 
     :param dev: Path for the already populated disk device.
     :raises InstanceDeployFailure: Image is not UEFI bootable.
     :returns: The UUID of the partition.
     """
+    warnings.warn("get_uefi_disk_identifier is deprecated, use "
+                  "find_efi_partition and get_partition_information instead",
+                  DeprecationWarning)
+
     partition_id = None
     try:
         report, _ = utils.execute('fdisk', '-l', dev, run_as_root=True)
@@ -194,11 +265,20 @@ def get_uefi_disk_identifier(dev):
     return disk_identifier
 
 
-def is_iscsi_device(dev, node_uuid):
-    """check whether the device path belongs to an iscsi device. """
+_ISCSI_PREFIX = "iqn.2008-10.org.openstack:"
 
-    iscsi_id = "iqn.2008-10.org.openstack:%s" % node_uuid
-    return iscsi_id in dev
+
+# TODO(dtantsur): deprecate node_uuid here, it's not overly useful (any iSCSI
+# device should get the same treatment).
+def is_iscsi_device(dev, node_uuid=None):
+    """Check whether the device path belongs to an iSCSI device.
+
+    If node UUID is provided, checks that the device belongs to this UUID.
+    """
+    if node_uuid:
+        return (_ISCSI_PREFIX + node_uuid) in dev
+    else:
+        return _ISCSI_PREFIX in dev
 
 
 def is_last_char_digit(dev):
@@ -208,6 +288,27 @@ def is_last_char_digit(dev):
     return False
 
 
+def partition_index_to_path(device, index):
+    """Guess a partition path based on its device and index.
+
+    :param device: Device path.
+    :param index: Partition index.
+    """
+    # the actual device names in the baremetal are like /dev/sda, /dev/sdb etc.
+    # While for the iSCSI device, the naming convention has a format which has
+    # iqn also embedded in it.
+    # When this function is called by ironic-conductor, the iSCSI device name
+    # should be appended by "part%d". While on the baremetal, it should name
+    # the device partitions as /dev/sda1 and not /dev/sda-part1.
+    if is_iscsi_device(device):
+        part_template = '%s-part%d'
+    elif is_last_char_digit(device):
+        part_template = '%sp%d'
+    else:
+        part_template = '%s%d'
+    return part_template % (device, index)
+
+
 def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
                     configdrive_mb, node_uuid, commit=True,
                     boot_option="netboot", boot_mode="bios",
@@ -245,18 +346,6 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
     LOG.debug("Starting to partition the disk device: %(dev)s "
               "for node %(node)s",
               {'dev': dev, 'node': node_uuid})
-    # the actual device names in the baremetal are like /dev/sda, /dev/sdb etc.
-    # While for the iSCSI device, the naming convention has a format which has
-    # iqn also embedded in it.
-    # When this function is called by ironic-conductor, the iSCSI device name
-    # should be appended by "part%d". While on the baremetal, it should name
-    # the device partitions as /dev/sda1 and not /dev/sda-part1.
-    if is_iscsi_device(dev, node_uuid):
-        part_template = dev + '-part%d'
-    elif is_last_char_digit(dev):
-        part_template = dev + 'p%d'
-    else:
-        part_template = dev + '%d'
     part_dict = {}
 
     if disk_label is None:
@@ -270,13 +359,15 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
         part_num = dp.add_partition(CONF.disk_utils.efi_system_partition_size,
                                     fs_type='fat32',
                                     boot_flag='boot')
-        part_dict['efi system partition'] = part_template % part_num
+        part_dict['efi system partition'] = partition_index_to_path(
+            dev, part_num)
 
     if (boot_mode == "bios" and boot_option == "local" and disk_label == "gpt"
         and not cpu_arch.startswith('ppc64')):
         part_num = dp.add_partition(CONF.disk_utils.bios_boot_partition_size,
                                     boot_flag='bios_grub')
-        part_dict['BIOS Boot partition'] = part_template % part_num
+        part_dict['BIOS Boot partition'] = partition_index_to_path(
+            dev, part_num)
 
     # NOTE(mjturek): With ppc64* nodes, partition images are expected to have
     # a PrEP partition at the start of the disk. This is an 8 MiB partition
@@ -290,25 +381,26 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
         boot_flag = 'boot' if disk_label == 'msdos' else None
         part_num = dp.add_partition(8, part_type='primary',
                                     boot_flag=boot_flag, extra_flags=['prep'])
-        part_dict['PReP Boot partition'] = part_template % part_num
+        part_dict['PReP Boot partition'] = partition_index_to_path(
+            dev, part_num)
     if ephemeral_mb:
         LOG.debug("Add ephemeral partition (%(size)d MB) to device: %(dev)s "
                   "for node %(node)s",
                   {'dev': dev, 'size': ephemeral_mb, 'node': node_uuid})
         part_num = dp.add_partition(ephemeral_mb)
-        part_dict['ephemeral'] = part_template % part_num
+        part_dict['ephemeral'] = partition_index_to_path(dev, part_num)
     if swap_mb:
         LOG.debug("Add Swap partition (%(size)d MB) to device: %(dev)s "
                   "for node %(node)s",
                   {'dev': dev, 'size': swap_mb, 'node': node_uuid})
         part_num = dp.add_partition(swap_mb, fs_type='linux-swap')
-        part_dict['swap'] = part_template % part_num
+        part_dict['swap'] = partition_index_to_path(dev, part_num)
     if configdrive_mb:
         LOG.debug("Add config drive partition (%(size)d MB) to device: "
                   "%(dev)s for node %(node)s",
                   {'dev': dev, 'size': configdrive_mb, 'node': node_uuid})
         part_num = dp.add_partition(configdrive_mb)
-        part_dict['configdrive'] = part_template % part_num
+        part_dict['configdrive'] = partition_index_to_path(dev, part_num)
 
     # NOTE(lucasagomes): Make the root partition the last partition. This
     # enables tools like cloud-init's growroot utility to expand the root
@@ -324,12 +416,12 @@ def make_partitions(dev, root_mb, swap_mb, ephemeral_mb,
 
     part_num = dp.add_partition(root_mb, boot_flag=boot_val)
 
-    part_dict['root'] = part_template % part_num
+    part_dict['root'] = partition_index_to_path(dev, part_num)
 
     if commit:
         # write to the disk
         dp.commit()
-        _trigger_device_rescan(dev)
+        trigger_device_rescan(dev)
     return part_dict
 
 
@@ -371,14 +463,36 @@ def qemu_img_info(path):
 
     out, err = utils.execute('env', 'LC_ALL=C', 'LANG=C',
                              'qemu-img', 'info', path,
-                             prlimit=QEMU_IMG_LIMITS)
+                             prlimit=_qemu_img_limits())
     return imageutils.QemuImgInfo(out)
 
 
+def _retry_on_res_temp_unavailable(exc):
+    if (isinstance(exc, processutils.ProcessExecutionError)
+            and ('Resource temporarily unavailable' in exc.stderr
+                 or 'Cannot allocate memory' in exc.stderr)):
+        return True
+    return False
+
+
+@tenacity.retry(
+    retry=tenacity.retry_if_exception(_retry_on_res_temp_unavailable),
+    stop=tenacity.stop_after_attempt(CONF.disk_utils.image_convert_attempts),
+    reraise=True)
 def convert_image(source, dest, out_format, run_as_root=False):
     """Convert image to other format."""
     cmd = ('qemu-img', 'convert', '-O', out_format, source, dest)
-    utils.execute(*cmd, run_as_root=run_as_root, prlimit=QEMU_IMG_LIMITS)
+    try:
+        utils.execute(*cmd, run_as_root=run_as_root,
+                      prlimit=_qemu_img_limits(),
+                      use_standard_locale=True)
+    except processutils.ProcessExecutionError as e:
+        if ('Resource temporarily unavailable' in e.stderr
+            or 'Cannot allocate memory' in e.stderr):
+            LOG.debug('Failed to convert image, retrying. Error: %s', e)
+            # Sync disk caches before the next attempt
+            utils.execute('sync')
+        raise
 
 
 def populate_image(src, dst, conv_flags=None):
@@ -394,15 +508,14 @@ def block_uuid(dev):
 
     Try to fetch the UUID, if that fails, try to fetch the PARTUUID.
     """
-    out, _err = utils.execute('blkid', '-s', 'UUID', '-o', 'value', dev,
-                              run_as_root=True, check_exit_code=[0])
-    if not out:
+    info = get_device_information(dev, fields=['UUID', 'PARTUUID'])
+    if info.get('UUID'):
+        return info['UUID']
+    else:
         LOG.debug('Falling back to partition UUID as the block device UUID '
                   'was not found while examining %(device)s',
                   {'device': dev})
-        out, _err = utils.execute('blkid', '-s', 'PARTUUID', '-o', 'value',
-                                  dev, run_as_root=True, check_exit_code=[0])
-    return out.strip()
+        return info.get('PARTUUID', '')
 
 
 def get_image_mb(image_path, virtual_size=True):
@@ -422,7 +535,7 @@ def get_image_mb(image_path, virtual_size=True):
 def get_dev_block_size(dev):
     """Get the device size in 512 byte sectors."""
     block_sz, cmderr = utils.execute('blockdev', '--getsz', dev,
-                                     run_as_root=True, check_exit_code=[0])
+                                     run_as_root=True)
     return int(block_sz)
 
 
@@ -458,13 +571,24 @@ def destroy_disk_metadata(dev, node_uuid):
     # erasing partition data.
     # This is the same bug as
     # https://bugs.launchpad.net/ironic-python-agent/+bug/1737556
+
+    # Overwrite the Primary GPT, catch very small partitions (like EBRs)
     dd_device = 'of=%s' % dev
-    gpt_backup = get_dev_block_size(dev) - 33
-    dd_seek = 'seek=%i' % gpt_backup
-    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, 'count=33',
+    dd_count = 'count=%s' % GPT_SIZE_SECTORS
+    dev_size = get_dev_block_size(dev)
+    if dev_size < GPT_SIZE_SECTORS:
+        dd_count = 'count=%s' % dev_size
+    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, dd_count,
                   run_as_root=True, use_standard_locale=True)
-    utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, 'count=33',
-                  dd_seek, run_as_root=True, use_standard_locale=True)
+
+    # Overwrite the Secondary GPT, do this only if there could be one
+    if dev_size > GPT_SIZE_SECTORS:
+        gpt_backup = dev_size - GPT_SIZE_SECTORS
+        dd_seek = 'seek=%i' % gpt_backup
+        dd_count = 'count=%s' % GPT_SIZE_SECTORS
+        utils.execute('dd', 'bs=512', 'if=/dev/zero', dd_device, dd_count,
+                      dd_seek, run_as_root=True, use_standard_locale=True)
+
     # Go ahead and let sgdisk run as well.
     utils.execute('sgdisk', '-Z', dev, run_as_root=True,
                   use_standard_locale=True)
@@ -507,18 +631,34 @@ def _get_configdrive(configdrive, node_uuid, tempdir=None):
     else:
         data = configdrive
 
-    try:
-        data = io.BytesIO(base64.decode_as_bytes(data))
-    except TypeError:
-        error_msg = (_('Config drive for node %s is not base64 encoded '
-                       'or the content is malformed.') % node_uuid)
-        if is_url:
-            error_msg += _(' Downloaded from "%s".') % configdrive
-        raise exception.InstanceDeployFailure(error_msg)
-
     configdrive_file = tempfile.NamedTemporaryFile(delete=False,
                                                    prefix='configdrive',
                                                    dir=tempdir)
+
+    try:
+        data = io.BytesIO(base64.b64decode(data))
+    except Exception as exc:
+        if isinstance(data, bytes):
+            LOG.debug('Config drive for node %(node)s is not base64 encoded '
+                      '(%(error)s), assuming binary',
+                      {'node': node_uuid, 'error': exc})
+            configdrive_mb = int(math.ceil(len(data) / units.Mi))
+            configdrive_file.write(data)
+            configdrive_file.close()
+            return (configdrive_mb, configdrive_file.name)
+        else:
+            configdrive_file.close()
+            utils.unlink_without_raise(configdrive_file.name)
+
+            error_msg = (_('Config drive for node %(node)s is not base64 '
+                           'encoded or the content is malformed. '
+                           '%(cls)s: %(err)s.')
+                         % {'node': node_uuid, 'err': exc,
+                            'cls': type(exc).__name__})
+            if is_url:
+                error_msg += _(' Downloaded from "%s".') % configdrive
+            raise exception.InstanceDeployFailure(error_msg)
+
     configdrive_mb = 0
     with gzip.GzipFile('configdrive', 'rb', fileobj=data) as gunzipped:
         try:
@@ -736,11 +876,8 @@ def _get_labelled_partition(device_path, label, node_uuid):
     :returns: block device file for partition if it exists; otherwise it
               returns None.
     """
+    partprobe(device_path)
     try:
-        utils.execute('partprobe', device_path, run_as_root=True,
-                      attempts=CONF.disk_utils.partprobe_attempts)
-
-        # lsblk command
         output, err = utils.execute('lsblk', '-Po', 'name,label', device_path,
                                     check_exit_code=[0, 1],
                                     use_standard_locale=True, run_as_root=True)
@@ -755,11 +892,7 @@ def _get_labelled_partition(device_path, label, node_uuid):
 
     found_part = None
     if output:
-        for device in output.split('\n'):
-            dev = {key: value for key, value in (v.split('=', 1)
-                   for v in shlex.split(device))}
-            if not dev:
-                continue
+        for dev in utils.parse_device_tags(output):
             if dev['LABEL'].upper() == label.upper():
                 if found_part:
                     found_2 = '/dev/%(part)s' % {'part': dev['NAME'].strip()}
@@ -775,31 +908,6 @@ def _get_labelled_partition(device_path, label, node_uuid):
     return found_part
 
 
-def _is_disk_gpt_partitioned(device, node_uuid):
-    """Checks if the disk is GPT partitioned
-
-    :param device: The device path.
-    :param node_uuid: UUID of the Node. Used for logging.
-    :raises: InstanceDeployFailure, if any disk partitioning related
-        commands fail.
-    :param node_uuid: UUID of the Node
-    :returns: Boolean. Returns True if disk is GPT partitioned
-    """
-    try:
-        stdout, _stderr = utils.execute(
-            'blkid', '-p', '-o', 'value', '-s', 'PTTYPE', device,
-            use_standard_locale=True, run_as_root=True)
-    except (processutils.UnknownArgumentError,
-            processutils.ProcessExecutionError, OSError) as e:
-        msg = (_('Failed to retrieve partition table type for disk %(disk)s '
-                 'for node %(node)s. Error: %(error)s') %
-               {'disk': device, 'node': node_uuid, 'error': e})
-        LOG.error(msg)
-        raise exception.InstanceDeployFailure(msg)
-
-    return (stdout.lower().strip() == 'gpt')
-
-
 def _fix_gpt_structs(device, node_uuid):
     """Checks backup GPT data structures and moves them to end of the device
 
@@ -834,7 +942,7 @@ def fix_gpt_partition(device, node_uuid):
     :raises: InstanceDeployFailure if exception is caught.
     """
     try:
-        disk_is_gpt_partitioned = _is_disk_gpt_partitioned(device, node_uuid)
+        disk_is_gpt_partitioned = (get_partition_table_type(device) == 'gpt')
         if disk_is_gpt_partitioned:
             _fix_gpt_structs(device, node_uuid)
     except Exception as e:
@@ -845,7 +953,50 @@ def fix_gpt_partition(device, node_uuid):
         raise exception.InstanceDeployFailure(msg)
 
 
-def _trigger_device_rescan(device):
+def udev_settle():
+    """Wait for the udev event queue to settle.
+
+    Wait for the udev event queue to settle to make sure all devices
+    are detected once the machine boots up.
+
+    :return: True on success, False otherwise.
+    """
+    LOG.debug('Waiting until udev event queue is empty')
+    try:
+        utils.execute('udevadm', 'settle')
+    except processutils.ProcessExecutionError as e:
+        LOG.warning('Something went wrong when waiting for udev '
+                    'to settle. Error: %s', e)
+        return False
+    else:
+        return True
+
+
+def partprobe(device, attempts=None):
+    """Probe partitions on the given device.
+
+    :param device: The block device containing paritions that is attempting
+                   to be updated.
+    :param attempts: Number of attempts to run partprobe, the default is read
+                     from the configuration.
+    :return: True on success, False otherwise.
+    """
+    if attempts is None:
+        attempts = CONF.disk_utils.partprobe_attempts
+
+    try:
+        utils.execute('partprobe', device, run_as_root=True, attempts=attempts)
+    except (processutils.UnknownArgumentError,
+            processutils.ProcessExecutionError, OSError) as e:
+        LOG.warning("Unable to probe for partitions on device %(device)s, "
+                    "the partitioning table may be broken. Error: %(error)s",
+                    {'device': device, 'error': e})
+        return False
+    else:
+        return True
+
+
+def trigger_device_rescan(device, attempts=None):
     """Sync and trigger device rescan.
 
     Disk partition performed via parted, when performed on a ramdisk
@@ -863,25 +1014,25 @@ def _trigger_device_rescan(device):
 
     :param device: The block device containing paritions that is attempting
                    to be updated.
+    :param attempts: Number of attempts to run partprobe, the default is read
+                     from the configuration.
+    :return: True on success, False otherwise.
     """
-    # TODO(TheJulia): This helper method was broken out for re-use, and
-    # does not have an explicit test case for itself, and it should, but
-    # it's steps are fairly explicitly checked with mocks on execute in
-    # the partition code.
-    LOG.debug('Explicitly calling sync to force buffer/cache flush.')
+    LOG.debug('Explicitly calling sync to force buffer/cache flush')
     utils.execute('sync')
     # Make sure any additions to the partitioning are reflected in the
     # kernel.
-    LOG.debug('Waiting until udev event queue is empty')
-    utils.execute('udevadm', 'settle')
+    udev_settle()
+    partprobe(device, attempts=attempts)
     try:
-        utils.execute('partprobe', device, run_as_root=True,
-                      attempts=CONF.disk_utils.partprobe_attempts)
         # Also verify that the partitioning is correct now.
         utils.execute('sgdisk', '-v', device, run_as_root=True)
     except processutils.ProcessExecutionError as exc:
-        LOG.warning('Failed to verify partitioning after creating '
-                    'partition(s): %s', exc)
+        LOG.warning('Failed to verify partition tables on device %(dev)s: '
+                    '%(err)s', {'dev': device, 'err': exc})
+        return False
+    else:
+        return True
 
 
 def create_config_drive_partition(node_uuid, device, configdrive):
@@ -924,7 +1075,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
         else:
             cur_parts = set(part['number'] for part in list_partitions(device))
 
-            if _is_disk_gpt_partitioned(device, node_uuid):
+            if get_partition_table_type(device) == 'gpt':
                 create_option = '0:-%dMB:0' % MAX_CONFIG_DRIVE_SIZE_MB
                 utils.execute('sgdisk', '-n', create_option, device,
                               run_as_root=True)
@@ -967,7 +1118,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                               'mkpart', 'primary', 'fat32', startlimit,
                               endlimit, run_as_root=True)
             # Trigger device rescan
-            _trigger_device_rescan(device)
+            trigger_device_rescan(device)
 
             upd_parts = set(part['number'] for part in list_partitions(device))
             new_part = set(upd_parts) - set(cur_parts)
@@ -977,15 +1128,9 @@ def create_config_drive_partition(node_uuid, device, configdrive):
                       'Unable to retrieve config drive partition information.')
                     % {'device': device})
 
-            if is_iscsi_device(device, node_uuid):
-                config_drive_part = '%s-part%s' % (device, new_part.pop())
-            elif is_last_char_digit(device):
-                config_drive_part = '%sp%s' % (device, new_part.pop())
-            else:
-                config_drive_part = '%s%s' % (device, new_part.pop())
+            config_drive_part = partition_index_to_path(device, new_part.pop())
 
-            LOG.debug('Waiting until udev event queue is empty')
-            utils.execute('udevadm', 'settle')
+            udev_settle()
 
             # NOTE(vsaienko): check that devise actually exists,
             # it is not handled by udevadm when using ISCSI, for more info see:
@@ -994,8 +1139,7 @@ def create_config_drive_partition(node_uuid, device, configdrive):
             LOG.debug('Waiting for the config drive partition %(part)s '
                       'on node %(node)s to be ready for writing.',
                       {'part': config_drive_part, 'node': node_uuid})
-            utils.execute('test', '-e', config_drive_part,
-                          check_exit_code=[0], attempts=15,
+            utils.execute('test', '-e', config_drive_part, attempts=15,
                           delay_on_retry=True)
 
         dd(confdrive_file, config_drive_part)
