@@ -18,21 +18,25 @@
 
 """Utilities and helper functions."""
 
+import contextlib
 import copy
 import errno
 import ipaddress
 import logging
 import os
 import re
+import shlex
+import shutil
+import tempfile
 from urllib import parse as urlparse
 
 from oslo_concurrency import processutils
 from oslo_config import cfg
-from oslo_service import loopingcall
 from oslo_utils import excutils
 from oslo_utils import specs_matcher
 from oslo_utils import strutils
 from oslo_utils import units
+import tenacity
 
 from ironic_lib.common.i18n import _
 from ironic_lib import exception
@@ -60,34 +64,28 @@ VALID_ROOT_DEVICE_HINTS = {
 ROOT_DEVICE_HINTS_GRAMMAR = specs_matcher.make_grammar()
 
 
-def execute(*cmd, **kwargs):
+def execute(*cmd, use_standard_locale=False, log_stdout=True, **kwargs):
     """Convenience wrapper around oslo's execute() method.
 
     Executes and logs results from a system command. See docs for
     oslo_concurrency.processutils.execute for usage.
 
-    :param \*cmd: positional arguments to pass to processutils.execute()
-    :param use_standard_locale: keyword-only argument. True | False.
-                                Defaults to False. If set to True,
+    :param cmd: positional arguments to pass to processutils.execute()
+    :param use_standard_locale: Defaults to False. If set to True,
                                 execute command with standard locale
                                 added to environment variables.
-    :param log_stdout: keyword-only argument. True | False. Defaults
-                       to True. If set to True, logs the output.
-    :param \*\*kwargs: keyword arguments to pass to processutils.execute()
+    :param log_stdout: Defaults to True. If set to True, logs the output.
+    :param kwargs: keyword arguments to pass to processutils.execute()
     :returns: (stdout, stderr) from process execution
     :raises: UnknownArgumentError on receiving unknown arguments
     :raises: ProcessExecutionError
     :raises: OSError
     """
-
-    use_standard_locale = kwargs.pop('use_standard_locale', False)
     if use_standard_locale:
         env = kwargs.pop('env_variables', os.environ.copy())
         env['LC_ALL'] = 'C'
         kwargs['env_variables'] = env
 
-    log_stdout = kwargs.pop('log_stdout', True)
-
     # If root_helper config is not specified, no commands are run as root.
     run_as_root = kwargs.get('run_as_root', False)
     if run_as_root:
@@ -105,6 +103,26 @@ def execute(*cmd, **kwargs):
     return result
 
 
+def try_execute(*cmd, **kwargs):
+    """The same as execute but returns None on error.
+
+    Executes and logs results from a system command. See docs for
+    oslo_concurrency.processutils.execute for usage.
+
+    Instead of raising an exception on failure, this method simply
+    returns None in case of failure.
+
+    :param cmd: positional arguments to pass to processutils.execute()
+    :param kwargs: keyword arguments to pass to processutils.execute()
+    :raises: UnknownArgumentError on receiving unknown arguments
+    :returns: tuple of (stdout, stderr) or None in some error cases
+    """
+    try:
+        return execute(*cmd, **kwargs)
+    except (processutils.ProcessExecutionError, OSError) as e:
+        LOG.debug('Command failed: %s', e)
+
+
 def mkfs(fs, path, label=None):
     """Format a file or block device
 
@@ -165,7 +183,7 @@ def dd(src, dst, *args):
     """
     LOG.debug("Starting dd process.")
     execute('dd', 'if=%s' % src, 'of=%s' % dst, *args,
-            use_standard_locale=True, run_as_root=True, check_exit_code=[0])
+            use_standard_locale=True, run_as_root=True)
 
 
 def is_http_url(url):
@@ -326,6 +344,101 @@ def parse_root_device_hints(root_device):
     return _append_operator_to_hints(root_device)
 
 
+def find_devices_by_hints(devices, root_device_hints):
+    """Find all devices that match the root device hints.
+
+    Try to find devices that match the root device hints. In order
+    for a device to be matched it needs to satisfy all the given hints.
+
+    :param devices: A list of dictionaries representing the devices
+                    containing one or more of the following keys:
+
+        :name: (String) The device name, e.g /dev/sda
+        :size: (Integer) Size of the device in *bytes*
+        :model: (String) Device model
+        :vendor: (String) Device vendor name
+        :serial: (String) Device serial number
+        :wwn: (String) Unique storage identifier
+        :wwn_with_extension: (String): Unique storage identifier with
+                             the vendor extension appended
+        :wwn_vendor_extension: (String): United vendor storage identifier
+        :rotational: (Boolean) Whether it's a rotational device or
+                     not. Useful to distinguish HDDs (rotational) and SSDs
+                     (not rotational).
+        :hctl: (String): The SCSI address: Host, channel, target and lun.
+                         For example: '1:0:0:0'.
+        :by_path: (String): The alternative device name,
+                  e.g. /dev/disk/by-path/pci-0000:00
+
+    :param root_device_hints: A dictionary with the root device hints.
+    :raises: ValueError, if some information is invalid.
+    :returns: A generator with all matching devices as dictionaries.
+    """
+    LOG.debug('Trying to find devices from "%(devs)s" that match the '
+              'device hints "%(hints)s"',
+              {'devs': ', '.join([d.get('name') for d in devices]),
+               'hints': root_device_hints})
+    parsed_hints = parse_root_device_hints(root_device_hints)
+    for dev in devices:
+        device_name = dev.get('name')
+
+        for hint in parsed_hints:
+            hint_type = VALID_ROOT_DEVICE_HINTS[hint]
+            device_value = dev.get(hint)
+            hint_value = parsed_hints[hint]
+
+            if hint_type is str:
+                try:
+                    device_value = _normalize_hint_expression(device_value,
+                                                              hint)
+                except ValueError:
+                    LOG.warning(
+                        'The attribute "%(attr)s" of the device "%(dev)s" '
+                        'has an empty value. Skipping device.',
+                        {'attr': hint, 'dev': device_name})
+                    break
+
+            if hint == 'size':
+                # Since we don't support units yet we expect the size
+                # in GiB for now
+                device_value = device_value / units.Gi
+
+            LOG.debug('Trying to match the device hint "%(hint)s" '
+                      'with a value of "%(hint_value)s" against the same '
+                      'device\'s (%(dev)s) attribute with a value of '
+                      '"%(dev_value)s"', {'hint': hint, 'dev': device_name,
+                                          'hint_value': hint_value,
+                                          'dev_value': device_value})
+
+            # NOTE(lucasagomes): Boolean hints are not supported by
+            # specs_matcher.match(), so we need to do the comparison
+            # ourselves
+            if hint_type is bool:
+                try:
+                    device_value = strutils.bool_from_string(device_value,
+                                                             strict=True)
+                except ValueError:
+                    LOG.warning('The attribute "%(attr)s" (with value '
+                                '"%(value)s") of device "%(dev)s" is not '
+                                'a valid Boolean. Skipping device.',
+                                {'attr': hint, 'value': device_value,
+                                 'dev': device_name})
+                    break
+                if device_value == hint_value:
+                    continue
+
+            elif specs_matcher.match(device_value, hint_value):
+                continue
+
+            LOG.debug('The attribute "%(attr)s" (with value "%(value)s") '
+                      'of device "%(dev)s" does not match the hint %(hint)s',
+                      {'attr': hint, 'value': device_value,
+                       'dev': device_name, 'hint': hint_value})
+            break
+        else:
+            yield dev
+
+
 def match_root_device_hints(devices, root_device_hints):
     """Try to find a device that matches the root device hints.
 
@@ -356,68 +469,15 @@ def match_root_device_hints(devices, root_device_hints):
     :raises: ValueError, if some information is invalid.
     :returns: The first device to match all the hints or None.
     """
-    LOG.debug('Trying to find a device from "%(devs)s" that matches the '
-              'root device hints "%(hints)s"',
-              {'devs': ', '.join([d.get('name') for d in devices]),
-               'hints': root_device_hints})
-    parsed_hints = parse_root_device_hints(root_device_hints)
-    for dev in devices:
-        device_name = dev.get('name')
-
-        for hint in parsed_hints:
-            hint_type = VALID_ROOT_DEVICE_HINTS[hint]
-            device_value = dev.get(hint)
-            hint_value = parsed_hints[hint]
-
-            if hint_type is str:
-                try:
-                    device_value = _normalize_hint_expression(device_value,
-                                                              hint)
-                except ValueError:
-                    LOG.warning(
-                        'The attribute "%(attr)s" of the device "%(dev)s" '
-                        'has an empty value. Skipping device.',
-                        {'attr': hint, 'dev': device_name})
-                    break
-
-            if hint == 'size':
-                # Since we don't support units yet we expect the size
-                # in GiB for now
-                device_value = device_value / units.Gi
-
-            LOG.debug('Trying to match the root device hint "%(hint)s" '
-                      'with a value of "%(hint_value)s" against the same '
-                      'device\'s (%(dev)s) attribute with a value of '
-                      '"%(dev_value)s"', {'hint': hint, 'dev': device_name,
-                                          'hint_value': hint_value,
-                                          'dev_value': device_value})
-
-            # NOTE(lucasagomes): Boolean hints are not supported by
-            # specs_matcher.match(), so we need to do the comparison
-            # ourselves
-            if hint_type is bool:
-                try:
-                    device_value = strutils.bool_from_string(device_value,
-                                                             strict=True)
-                except ValueError:
-                    LOG.warning('The attribute "%(attr)s" (with value '
-                                '"%(value)s") of device "%(dev)s" is not '
-                                'a valid Boolean. Skipping device.',
-                                {'attr': hint, 'value': device_value,
-                                 'dev': device_name})
-                    break
-                if device_value == hint_value:
-                    continue
-                break
-
-            if not specs_matcher.match(device_value, hint_value):
-                break
-        else:
-            LOG.info('Device found! The device "%s" matches the root '
-                     'device hints', device_name)
-            return dev
-
-    LOG.warning('No device found that matches the root device hints')
+    try:
+        dev = next(find_devices_by_hints(devices, root_device_hints))
+    except StopIteration:
+        LOG.warning('No device found that matches the root device hints %s',
+                    root_device_hints)
+    else:
+        LOG.info('Root device found! The device "%s" matches the root '
+                 'device hints %s', dev, root_device_hints)
+        return dev
 
 
 def wait_for_disk_to_become_available(device):
@@ -435,21 +495,16 @@ def wait_for_disk_to_become_available(device):
     :raises: IronicException If the disk fails to become
         available.
     """
-    retries = [0]
     pids = ['']
     stderr = ['']
     interval = CONF.disk_partitioner.check_device_interval
     max_retries = CONF.disk_partitioner.check_device_max_retries
 
-    def _wait_for_disk(device, retries, max_retries, pids, stderr):
+    def _wait_for_disk():
         # A regex is likely overkill here, but variations in fuser
         # means we should likely use it.
         fuser_pids_re = re.compile(r'\d+')
 
-        retries[0] += 1
-        if retries[0] > max_retries:
-            raise loopingcall.LoopingCallDone()
-
         # There are 'psmisc' and 'busybox' versions of the 'fuser' program. The
         # 'fuser' programs differ in how they output data to stderr.  The
         # busybox version does not output the filename to stderr, while the
@@ -474,7 +529,7 @@ def wait_for_disk_to_become_available(device):
                                run_as_root=True)
 
             if not out and not err:
-                raise loopingcall.LoopingCallDone()
+                return True
 
             stderr[0] = err
             # NOTE: findall() returns a list of matches, or an empty list if no
@@ -484,13 +539,16 @@ def wait_for_disk_to_become_available(device):
         except processutils.ProcessExecutionError as exc:
             LOG.warning('Failed to check the device %(device)s with fuser:'
                         ' %(err)s', {'device': device, 'err': exc})
+        return False
 
-    timer = loopingcall.FixedIntervalLoopingCall(
-        _wait_for_disk,
-        device, retries, max_retries, pids, stderr)
-    timer.start(interval=interval).wait()
-
-    if retries[0] > max_retries:
+    retry = tenacity.retry(
+        retry=tenacity.retry_if_result(lambda r: not r),
+        stop=tenacity.stop_after_attempt(max_retries),
+        wait=tenacity.wait_fixed(interval),
+        reraise=True)
+    try:
+        retry(_wait_for_disk)()
+    except tenacity.RetryError:
         if pids[0]:
             raise exception.IronicException(
                 _('Processes with the following PIDs are holding '
@@ -525,3 +583,76 @@ def get_route_source(dest, ignore_link_local=True):
     except (IndexError, ValueError):
         LOG.debug('No route to host %(dest)s, route record: %(rec)s',
                   {'dest': dest, 'rec': out})
+
+
+@contextlib.contextmanager
+def mounted(source, dest=None, opts=None, fs_type=None,
+            mount_attempts=1, umount_attempts=3):
+    """A context manager for a temporary mount.
+
+    :param source: A device to mount.
+    :param dest: Mount destination. If not specified, a temporary directory
+        will be created and removed afterwards. An existing destination is
+        not removed.
+    :param opts: Mount options (``-o`` argument).
+    :param fs_type: File system type (``-t`` argument).
+    :param mount_attempts: A number of attempts to mount the device.
+    :param umount_attempts: A number of attempts to unmount the device.
+    :returns: A generator yielding the destination.
+    """
+    params = []
+    if opts:
+        params.extend(['-o', ','.join(opts)])
+    if fs_type:
+        params.extend(['-t', fs_type])
+
+    if dest is None:
+        dest = tempfile.mkdtemp()
+        clean_up = True
+    else:
+        clean_up = False
+
+    mounted = False
+    try:
+        execute("mount", source, dest, *params, run_as_root=True,
+                attempts=mount_attempts, delay_on_retry=True)
+        mounted = True
+        yield dest
+    finally:
+        if mounted:
+            try:
+                execute("umount", dest, run_as_root=True,
+                        attempts=umount_attempts, delay_on_retry=True)
+            except (EnvironmentError,
+                    processutils.ProcessExecutionError) as exc:
+                LOG.warning(
+                    'Unable to unmount temporary location %(dest)s: %(err)s',
+                    {'dest': dest, 'err': exc})
+                # NOTE(dtantsur): don't try to remove a still mounted location
+                clean_up = False
+
+        if clean_up:
+            try:
+                shutil.rmtree(dest)
+            except EnvironmentError as exc:
+                LOG.warning(
+                    'Unable to remove temporary location %(dest)s: %(err)s',
+                    {'dest': dest, 'err': exc})
+
+
+def parse_device_tags(output):
+    """Parse tags from the lsblk/blkid output.
+
+    Parses format KEY="VALUE" KEY2="VALUE2".
+
+    :return: a generator yielding dicts with information from each line.
+    """
+    for line in output.strip().split('\n'):
+        if line.strip():
+            try:
+                yield {key: value for key, value in
+                       (v.split('=', 1) for v in shlex.split(line))}
+            except ValueError as err:
+                raise ValueError(
+                    _("Malformed blkid/lsblk output line '%(line)s': %(err)s")
+                    % {'line': line, 'err': err})
